#include "GS_Problem.h"
#include "ColorHelper.h"
#include "ImageProcessing.h"
#include "ImageIO.h"
#include "GSF_NPR.h"

void GS_Problem::PreProcessInput()
{
	preProcessInput(this->input.img);
	//bug
	preProcessInput(this->input.data);
}

void GS_Problem::preProcessInput(GS_Image& img)
{
	ChannelsType orgChannelsType = img.channelsType;
	GS_Image imgToFlt;
	if(this->input.img.Shape().nBands == 3)
		imgToFlt = ColorHelper::ConvertChannels(CT_RGB, this->input.img);
	else
		imgToFlt = ColorHelper::ConvertChannels(CT_LUM, this->input.img);

	FilterAuxData auxData;
	auxData.emData = this->input.emData;
	auxData.imgFN  = this->input.imgFN;

	FilterImg(imgToFlt, this->params.preprocessFilters, auxData);
	img = ColorHelper::ConvertChannels(orgChannelsType, imgToFlt);
}

void GS_Problem::FilterImg(GS_Image& img, const vector<Filter>& filters, const FilterAuxData& auxData)
{
	if(filters.size() > 0)
	{
		ChannelsType orgChannelsType = img.channelsType;
		INSIST((orgChannelsType == CT_RGB) || (orgChannelsType == CT_LUM));

		GS_Image orgImg;

		GSF_NPR::Params nprParams = GSF_NPR::Params::Default_EdgeVis();
		for(vector<Filter>::const_iterator currFilter = filters.begin(); 
			currFilter != filters.end();
			currFilter++)
		{
			GenericOpts fltParams = currFilter->params;
			
			switch(currFilter->type)
			{
				case PFT_DENOISE:
					INSIST(fltParams.floatList.size() == 2);
					img = ImageProcessing::BilateralFilter(img, fltParams.floatList[0], fltParams.floatList[1]);
					break;

				case PFT_SATURATE_COLOR:
					INSIST(fltParams.floatList.size() == 1);
					img = ImageProcessing::ModifyColorSat(img, fltParams.floatList[0]);
					break;

				case PFT_EDGE_VIS1:
					INSIST(fltParams.floatList.size() == 1);
					nprParams.edgeLenNorm = fltParams.floatList[0];
					img = GSF_NPR::VisBrushStrokes1(img, auxData.emData, nprParams);
					break;

				case PFT_LUM2RGB:
					INSIST(orgChannelsType == CT_LUM);
					INSIST(fltParams.floatList.size() == 1);
					ImageIO::ReadFile(orgImg, auxData.imgFN);					
					img = ColorHelper::NormalizeRGBbyLum(orgImg, img, fltParams.floatList[0]);
					break;

				case PFT_ROBUST_RESCALE:
					INSIST(fltParams.floatList.size() == 2);
					CHECK_RANGE(fltParams.floatList[0], 0.0f, fltParams.floatList[1]);
					CHECK_RANGE(fltParams.floatList[1], fltParams.floatList[0], 1.0f);
					ImageProcessing::Rescale(img, fltParams.floatList[0], fltParams.floatList[1], 0.0f, 1.0f);
					break;

				case PFT_EDGE_VIS2:
					INSIST(fltParams.floatList.size() == 1);
					nprParams.edgeLenNorm = fltParams.floatList[0];
					img = GSF_NPR::VisBrushStrokes2(img, auxData.emData, FVector3(0,0,0), nprParams);
					break;

				case PFT_SHARPEN:		
				default:
					REPORT_FAILURE("Unknown filter type - %i\n", currFilter->type);
					break;
			}
		}
	}
}


GS_Problem::Filter GS_Problem::Filter::Default(ProbFilterType _type)
{
	GS_Problem::Filter filter;
	filter.type = _type;

	switch(filter.type)
	{
		case PFT_SATURATE_COLOR:
			filter.params.floatList.push_back(1.1f); //saturate amount
			break;

		case PFT_SHARPEN:
			filter.params.floatList.push_back(1.0f); //sharpen amount
			break;

		case PFT_DENOISE:
			filter.params.floatList.push_back(3.0f);  //sigmaD
			filter.params.floatList.push_back(0.03f); //sigmaD
			break;

		case PFT_LUM2RGB:
			filter.params.floatList.push_back(0.3f);  //saturation
			break;

		case PFT_ROBUST_RESCALE:
			filter.params.floatList.push_back(0.01f);  //minPT
			filter.params.floatList.push_back(0.99f);  //maxPT
			break;

		case PFT_EDGE_VIS1:
			filter.params.floatList.push_back(1500.0f);  //edgeLenNorm
			break;
		case PFT_EDGE_VIS2:
			filter.params.floatList.push_back(1500.0f);  //edgeLenNorm
			break;

		default:
			REPORT_FAILURE("Unknown filter type - %i\n", filter.type);
			break;
	}

	return filter;
}

void GS_Problem::Filter::Write(ofstream &outStream, string paramPrefix) const
{
	CHECK_RANGE(this->type, 0, PFT_LAST);
	outStream << paramPrefix << "type = " 	
		<< ProbFilterTypeStr[this->type] << "\n";
}


void GS_Problem::Filter::Dump(string prefix) const
{
	CHECK_RANGE(this->type, 0, PFT_LAST);

	const char* prefixPtr = prefix.c_str();
	printf("\n");
	printf("%s filter params:\n", prefixPtr);
	printf("%s   type - %s\n", prefixPtr, ProbFilterTypeStr[this->type]);
	this->params.Dump(prefix + "   ");
	printf("\n");
}

