#include "GS_VSolver.h"
#include "Disk.h"
#include "ImageIO.h"
#include "FlowHelper.h"

void GS_VSolver::CombineResults()
{
	REPORT("Combing results...\n");

	GS_Image solImg(this->frameShape);

	INSIST(this->frameCount > 0);
	for(int iFrame = 0; iFrame < this->frameCount; iFrame++)
	{
		printf("."); fflush(stdout);

		int defChannelIndex = 0;
		if(this->params.channelsToSolve.size() > 0)
			defChannelIndex = *(this->params.channelsToSolve.begin());

		string frameProbID = getProbID(this->params.imgFNs[0], defChannelIndex);
		GS_Problem::Params frameProbParam = GS_Problem::Params::Load(frameProbID);

		ImageIO::ReadFile(solImg, this->params.imgFNs[iFrame]);
		solImg = ColorHelper::ConvertChannels(frameProbParam.channelsType, solImg);
		INSIST(solImg.Shape() == this->frameShape);

		INSIST(this->params.channelsToSolve.size() > 0);
		for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
			iChannel != this->params.channelsToSolve.end();
			iChannel++)
		{
			loadFrame(iFrame, *iChannel);

			GS_Image channelSol = this->currFrameProb.GetResult();
			INSIST(channelSol.Shape() == this->maskShape);

			ImageProcessing::CopyChannel(solImg, channelSol, *iChannel, 0);
		}

		string frameSolFN = getSolFN(this->params.imgFNs[iFrame]);
		if(solImg.Shape().nBands == 1)
			solImg = ColorHelper::ConvertChannels(CT_LUM, solImg);
		else
			solImg = ColorHelper::ConvertChannels(CT_RGB, solImg);

		EdgeMetrics emData;
		string emDataFN = getEdgeFN(this->params.imgFNs[iFrame]);
		if(Disk::DoesFileExist(emDataFN))
			emData.Load(emDataFN, this->frameShape);
		vector<GS_Problem::Filter> postprocessFilters = this->currFrameProb.GetParams().postprocessFilters;
		GS_Problem::FilterAuxData auxData;
		auxData.emData = emData;
		auxData.imgFN  = this->params.imgFNs[iFrame];
		//postprocessFilters.push_back(GS_Problem::Filter::Default(PFT_EDGE_VIS2));
		GS_Problem::FilterImg(solImg, postprocessFilters, auxData);
		ImageIO::WriteFile(solImg, frameSolFN);
	}
	printf("\n");
}

void GS_VSolver::loadFrameContext(int iFrame, int iChannel)
{
	CHECK_RANGE(iFrame,   0, this->frameCount - 1);
	CHECK_RANGE(iChannel, 0, this->frameShape.nBands - 1);

	GS_Problem nullProb;	

	if(iFrame > 0)
	{
		this->prevFrameProb  = this->currFrameProb;
		this->currFrameProb  = this->nextFrameProb;
		this->currToPrevFlow = this->nextToCurrFlow.Clone();
	}
	else
	{
		this->prevFrameProb = nullProb;
		this->currToPrevFlow.ClearPixels();

		string currFrameProbID = getProbID(this->params.imgFNs[iFrame], iChannel);
		this->currFrameProb    = GS_Problem::Load(currFrameProbID);
		INSIST(this->currFrameProb.GetParams().problemShape.imgShape == this->maskShape);

		if(this->params.useLeftConstraint)
		{
			if(this->params.useFlow)
			{
				string currToPrevFlowFN = getFlowFN(this->params.imgFNs[iFrame]);
				FlowHelper::ReadFlowField(currToPrevFlowFN, this->currToPrevFlow);
				INSIST(this->currToPrevFlow.Shape() == this->flowShape);
			}

			string prevFrameProbID = getProbID(this->params.leftConstraintImgFN, iChannel);
			this->prevFrameProb    = GS_Problem::Load(prevFrameProbID);
		}		
	}

	if(iFrame < this->frameCount - 1)
	{
		string nextFrameProbID = getProbID(this->params.imgFNs[iFrame + 1], iChannel);
		this->nextFrameProb    = GS_Problem::Load(nextFrameProbID);
		INSIST(this->nextFrameProb.GetParams().problemShape.imgShape == this->maskShape);

		if(this->params.useFlow)
		{
			string nextToCurrFlowFN = getFlowFN(this->params.imgFNs[iFrame + 1]);
			FlowHelper::ReadFlowField(nextToCurrFlowFN, this->nextToCurrFlow);
			INSIST(this->nextToCurrFlow.Shape() == this->flowShape);
		}
		else
		{
			this->nextToCurrFlow.ClearPixels();
		}
	}
	else
	{
		this->nextFrameProb = nullProb;
		this->nextToCurrFlow.ClearPixels();

		if(this->params.useRightConstraint)
		{
			if(this->params.useFlow)
			{
				string nextToCurrFlowFN = getFlowFN(this->params.rightConstraintImgFN);
				FlowHelper::ReadFlowField(nextToCurrFlowFN, this->nextToCurrFlow);
				INSIST(this->nextToCurrFlow.Shape() == this->flowShape);
			}

			string nextFrameProbID = getProbID(this->params.rightConstraintImgFN, iChannel);
			this->nextFrameProb    = GS_Problem::Load(nextFrameProbID);
		}
	}

	this->params.pinchFuncScale = ColorHelper::GetGradScale(this->currFrameProb.GetParams().channelsType);
}

string GS_VSolver::getFlowFN(string imgFN)
{
	string flowFN = Disk::GetFileNameID(imgFN) + ".flow";
	return flowFN;
}

string GS_VSolver::getEdgeFN(string imgFN)
{
	string edgeFN = Disk::GetFileNameID(imgFN) + ".edg";
	return edgeFN;
}

string GS_VSolver::getProbID(string imgFN, int iChannel)
{
	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(imgFN));
	string probID = this->params.outDir + CImage::GetChannelStr(iChannel) + "-" + fileID;
	return probID;
}

string GS_VSolver::getSolFN(string imgFN)
{
	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(imgFN));
	string probID = this->params.outDir + "final-" + fileID + ".tga";
	return probID;
}


string GS_VSolver::getEdgeVisFN(string imgFN)
{
	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(imgFN));
	string probID = this->params.outDir + "ev-" + fileID + ".tga";
	return probID;
}

void GS_VSolver::loadFrame(int iFrame, int iChannel)
{
	CHECK_RANGE(iFrame,   0, this->frameCount - 1);
	CHECK_RANGE(iChannel, 0, this->frameShape.nBands - 1);

	GS_Problem nullProb;

	this->prevFrameProb    = nullProb;
	string currFrameProbID = getProbID(this->params.imgFNs[iFrame], iChannel);
	this->currFrameProb    = GS_Problem::Load(currFrameProbID);
	this->nextFrameProb    = nullProb;

	INSIST(this->currFrameProb.GetParams().problemShape.imgShape == this->maskShape);
}

