#include "GS_VSolver.h"
#include "ImageIO.h"
#include "LinearSolverDefs.h"

GS_VSolver::GS_VSolver(const Params &_params)
{
	this->params = _params;	
	initGlobals();
}

void GS_VSolver::Solve()
{
	INSIST(this->params.channelsToSolve.size() > 0);
	for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
		iChannel != this->params.channelsToSolve.end();
		iChannel++)
	{
		solve(*iChannel);
	}

	if((int) this->params.channelsToSolve.size() == this->frameShape.nBands)
	{
		CombineResults();
	}
}

void GS_VSolver::solve(int iChannel)
{
	REPORT("\nSolving channel - %i\n", iChannel);

	initChannel(iChannel);

	int numOfVars = getVarsInProb();
	INSIST(numOfVars > 0);
	SparseSolverParams solverParams = SparseSolverParams::Default();
	solverParams.nonZeroInMatGuess = numOfVars * 7;
	solverParams.usePrecondHB = true;
	solverParams.maxLevelsInPrecond = this->params.maxLevelsInPrecond;
	for(int iFrame = 0; iFrame < this->frameCount; iFrame++)
	{
		solverParams.imgShapes.push_back(this->frameShape);
	}
	SparseSolver solver(numOfVars, solverParams);

	REPORT("Loading frame equations...\n");
	loadEqs(iChannel, solver);

	solver.Solve(this->params.accuracy, this->params.maxIters, true);

	REPORT("Writing frame solutions...\n");
	saveChannelSol(iChannel, solver);

	REPORT("Done solving channel - %i\n", iChannel);	
}

void GS_VSolver::loadEqs(int iChannel, SparseSolver& solver)
{	
	for(int iFrame = 0; iFrame < this->frameCount; iFrame++)
	{
		printf("."); fflush(stdout);
		loadFrameContext(iFrame, iChannel);

		VarCoefList* neighsInPrevFrame = NULL;
		VarCoefList* neighsInNextFrame = NULL;

		if((iFrame > 0) || (params.useLeftConstraint == true))
		{
			neighsInPrevFrame = new VarCoefList[this->pixelsInFrameChannel];
			ENSURE(neighsInPrevFrame != NULL);
			collectTemporalNeighbours(neighsInPrevFrame, 
									  this->currToPrevFlow, true,
									  CFloatImage(), false);
		}
		
		if((iFrame < this->frameCount - 1) || (params.useRightConstraint == true))
		{
			neighsInNextFrame = new VarCoefList[this->pixelsInFrameChannel];
			ENSURE(neighsInNextFrame != NULL);
			collectTemporalNeighbours(neighsInNextFrame, 
									  CFloatImage(), false,
									  this->nextToCurrFlow, true);
		}

		for(int iProbLevel = 0;	iProbLevel < this->currFrameProb.GetLevelCount(); iProbLevel++)
		{	
			defineEqs(iChannel, iProbLevel, iFrame, 
					  neighsInPrevFrame, neighsInNextFrame,
					  solver);
		}

		delete[] neighsInPrevFrame;
		delete[] neighsInNextFrame;
	}
	printf("\n");
}


void GS_VSolver::saveChannelSol(int iChannel, SparseSolver& solver)
{
	for(int iFrame = 0; iFrame < this->frameCount; iFrame++)
	{
		printf("."); fflush(stdout);
		loadFrame(iFrame, iChannel);
		
		Indices probLevelsToSolve;

		for(int iProbLevel = 0;	iProbLevel < this->currFrameProb.GetLevelCount(); iProbLevel++)
		{
			probLevelsToSolve.insert(iProbLevel);

			GS_ProblemLevel* problemLevel = this->currFrameProb.GetLevel(iProbLevel);
			GS_Image         levelFlt     = problemLevel->flt;

			uint nodeAddr = 0;
			uint varAddr  = getVarAddr(nodeAddr, iProbLevel, iFrame);
			for(int y = 0; y < this->frameShape.height; y++)
			for(int x = 0; x < this->frameShape.width;  x++, nodeAddr++, varAddr++)
			{
				double sol = solver.GetSolution(varAddr);
				levelFlt[nodeAddr] = (imgType) sol;
			}
		}

		this->currFrameProb.UpdateResult(probLevelsToSolve);

		if(this->params.writeProb || this->params.writeSol)
		{
			string probID = getProbID(this->params.imgFNs[iFrame], iChannel);
			if(this->params.writeProb)
			{
				this->currFrameProb.Write(probID, "raw", probLevelsToSolve);
			}
			this->currFrameProb.WriteSol(probID, "tga", false);
		}
	}
	printf("\n");
}


void GS_VSolver::defineEqs(int iChannel, int iProbLevel, int iFrame, 
						   VarCoefList* neighsInPrevFrame, VarCoefList* neighsInNextFrame,
						   SparseSolver& solver)
{
	CHECK_RANGE(iChannel, 0, this->frameShape.nBands - 1);

	const int EquationVars     = 5;
	const int dx[EquationVars] = { 0, -1,  0,  1,  0};
	const int dy[EquationVars] = {-1,  0,  0,  0,  1};

	int levelsToSolve = this->levelsInFrame[iFrame];
	INSIST(levelsToSolve == 1); // bug
	vector<size_v> eqVarAddr;
	vector<precise> eqVarCoef;
	eqVarAddr.reserve(100);
	eqVarCoef.reserve(100);
	
	EdgeMetrics emData;
	if(this->params.useSDI_weighing)
	{
		string emFN = getEdgeFN(this->params.imgFNs[iFrame]);
		emData.Load(emFN, this->frameShape);
	}

	GS_ProblemLevel* levelProb = this->currFrameProb.GetLevel(iProbLevel);
	GS_Image levelOrg,  levelFlt;
	GS_Image levelData, levelDataW;
	GS_Image levelOrgGF_X, levelOrgGF_Y;
	GS_Image levelFltGF_X, levelFltGF_Y;
	levelProb->GetData(levelOrg, levelFlt, 
					   levelData, levelDataW,
					   levelOrgGF_X, levelOrgGF_Y, levelFltGF_X, levelFltGF_Y);

	GS_Image dataImg, dataImgW, orgImg;
	if(this->params.useGlobalDatacost)
	{
		dataImg  =  this->currFrameProb.GetInputBuffer()->data;
		dataImgW =  this->currFrameProb.GetInputBuffer()->dataW;
		orgImg   =  this->currFrameProb.GetInputBuffer()->img;
	}
	else
	{
		dataImg    = levelData;
		levelDataW = levelDataW;
		orgImg     = levelOrg;
	}
	
	uint nodeAddr = 0;
	uint varAddr  = getVarAddr(nodeAddr, iProbLevel, iFrame);
	for(int y = 0; y < this->maskShape.height; y++)
	for(int x = 0; x < this->maskShape.width;  x++, nodeAddr++, varAddr++)
	{
		eqVarCoef.clear();
		eqVarAddr.clear();

		double rhs = 0;
		double fxy = levelFlt[nodeAddr];

		int    eqVarCount = 0;
		int    varIndex   = -1;
		double varCoef    = 0.0;
		for(int iEqVar = 0; iEqVar < EquationVars; iEqVar++)
		{
			if(iEqVar == 2)
			{
				INSIST(dx[iEqVar] == 0);
				INSIST(dy[iEqVar] == 0);

				eqVarAddr.push_back(varAddr);
				eqVarCoef.push_back(0.0f);
				varIndex = eqVarCount;
				eqVarCount++;

				varCoef += this->params.fidelity_data * dataImgW[nodeAddr];
				rhs     += this->params.fidelity_data * dataImgW[nodeAddr] * dataImg[nodeAddr];

				continue;
			}

			int neighbourX = x + dx[iEqVar];
			int neighbourY = y + dy[iEqVar];
			if(this->maskShape.InBounds(neighbourX, neighbourY) == false) continue;
	
			uint neighNodeAddr = this->maskShape.NodeIndex(neighbourX, neighbourY);
			uint neighVarAddr  = getVarAddr(neighNodeAddr, iProbLevel, iFrame);

			double neighPinch;
			double neighGrad;
			double neighCoef;

			if(neighbourY != y)
			{
				INSIST(neighbourX == x);
				if(y > neighbourY)
				{
					neighGrad  = (double) levelFltGF_Y[nodeAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[nodeAddr] - levelOrgGF_Y[nodeAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_Y[neighNodeAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[neighNodeAddr] - levelOrgGF_Y[neighNodeAddr]);
				}				
			}
			else
			{
				INSIST(neighbourX != x);
				if(x > neighbourX)
				{
					neighGrad  = (double) levelFltGF_X[nodeAddr];
					neighPinch = (double) myAbs(levelFltGF_X[nodeAddr] - levelOrgGF_X[nodeAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_X[neighNodeAddr];
					neighPinch = (double) myAbs(levelFltGF_X[neighNodeAddr] - levelOrgGF_X[neighNodeAddr]);
				}		
			}

			neighCoef = this->params.fidelity_sGrad / pow((this->params.pinchFuncScale * neighPinch) + 1.0, this->params.pinchFuncExp);

			if(this->params.useSDI_weighing)
			{
				double emWeight, emLen;
				if(neighbourY != y)
				{
					if(y > neighbourY)
					{
						emWeight = pow(sin(emData.edgeDirc[nodeAddr]), 2.0f);
						emLen    = emData.edgeLen[nodeAddr];

					}
					else
					{
						emWeight = pow(sin(emData.edgeDirc[neighNodeAddr]), 2.0f);
						emLen    = emData.edgeLen[neighNodeAddr];
					}
				}
				else
				{
					if(x > neighbourX)
					{
						emWeight = pow(cos(emData.edgeDirc[nodeAddr]), 2.0f);
						emLen    = emData.edgeLen[nodeAddr];

					}
					else
					{
						emWeight = pow(cos(emData.edgeDirc[neighNodeAddr]), 2.0f);
						emLen    = emData.edgeLen[neighNodeAddr];
					}			
				}

				emWeight *= emLen / this->params.sdiNorm;
				neighCoef /= (1.0 + emWeight);
			}

			if(neighCoef != 0.0f)
			{
				if(neighVarAddr >= varAddr)
				{
					eqVarAddr.push_back(neighVarAddr);
					eqVarCoef.push_back( precise(-neighCoef) );
					eqVarCount++;
				}

				rhs     += neighCoef * neighGrad;
				varCoef += neighCoef;
			}
		}

		if(this->params.useGlobalDatacost)
		{
			for(int vNeighLevel = iProbLevel + 1; vNeighLevel < levelsToSolve; vNeighLevel++)
			{
				uint vNeighVarAddr = getVarAddr(nodeAddr, vNeighLevel, iFrame);
				eqVarAddr.push_back(vNeighVarAddr);
				eqVarCoef.push_back( precise(this->params.fidelity_data * dataImgW[nodeAddr]) );
				eqVarCount++;
			}
		}

		for(int frameOffset = -1; frameOffset <= 1; frameOffset += 2)
		{	
			GS_Problem*  offsetFrameProb     = NULL;
			VarCoefList* neighsInOffsetFrame = NULL;			
			switch(frameOffset)
			{
			case -1:
				offsetFrameProb     = &this->prevFrameProb;
				neighsInOffsetFrame = neighsInPrevFrame;
				break;
			case 1:
				offsetFrameProb     = &this->nextFrameProb;
				neighsInOffsetFrame = neighsInNextFrame;
				break;
			default:
				REPORT_FAILURE("Unknown offset");
				break;
			}

			int offsetFrameIndex = iFrame + frameOffset;
			if ((offsetFrameIndex >= this->frameCount) && (params.useRightConstraint == false)) continue;
			if ((offsetFrameIndex < 0) && (params.useLeftConstraint == false)) continue;

			VarCoefList& currNodeNeighbours = neighsInOffsetFrame[nodeAddr];
			for(VarCoefList::const_iterator currNodeNeighbour = currNodeNeighbours.begin();
				currNodeNeighbour != currNodeNeighbours.end();
				currNodeNeighbour++)
			{					
				uint neighNodeAddr = currNodeNeighbour->first;
				double neighCoef   = currNodeNeighbour->second;

				if(neighCoef != 0.0f)
				{
					GS_Image offsetFrameOrgImg;
					GS_Image offsetFrameSolImg;
					if(this->params.useGlobalDatacost)
					{
						offsetFrameOrgImg  = offsetFrameProb->GetInputBuffer()->img;
						offsetFrameSolImg  = offsetFrameProb->GetResult();
					}
					else
					{
						offsetFrameOrgImg  = offsetFrameProb->GetLevel(iProbLevel)->org;
						offsetFrameSolImg  = offsetFrameProb->GetLevel(iProbLevel)->flt;
					}

					double tGrad = orgImg[nodeAddr] - offsetFrameOrgImg[neighNodeAddr];

					//bug
					//setTGrad_default(tGrad, neighCoef);
					setTGrad_temporalCoherence(tGrad, neighCoef);
					//setTGrad_deflicker(dataImg[nodeAddr], tGrad, neighCoef);
					//setTGrad_sharpenMotion(tGrad, neighCoef);

					rhs     += neighCoef * tGrad;
					varCoef += neighCoef;

					if(offsetFrameIndex < 0)
					{
						ENSURE(params.useLeftConstraint == true);
						rhs += offsetFrameSolImg[neighNodeAddr] * neighCoef;
						continue;
					}

					if(offsetFrameIndex >= this->frameCount)
					{
						ENSURE(params.useRightConstraint == true);						
						rhs += offsetFrameSolImg[neighNodeAddr] * neighCoef;
						continue;
					}

					if(frameOffset > 0)
					{
						//bug iProbLevel -> jProbLevel
						uint neighVarAddr = getVarAddr(neighNodeAddr, iProbLevel, offsetFrameIndex);
						INSIST(neighVarAddr > varAddr);
						eqVarAddr.push_back(neighVarAddr);
						eqVarCoef.push_back( precise(-neighCoef) );
						eqVarCount++;
					}
				}
			}
		}
		
		eqVarCoef[varIndex] = (precise) varCoef;
		ENSURE(varCoef > 0.0);
		solver.SetRow(varAddr, eqVarCount, (precise) fxy, (precise) rhs, eqVarAddr, eqVarCoef);
	}
}

