#include "SparseSymMat_CSR.h"

SparseSymMat_CSR::SparseSymMat_CSR(size_v _rows, size_v _cols, size_vv nonZeroCountGuess)
{
	INSIST(_rows > 0);
	INSIST(_cols > 0);
	INSIST(nonZeroCountGuess > 0);

	this->rowCount = _rows;
	this->colCount = _cols;
	this->nonZeroCount = nonZeroCountGuess;

	this->values  = NULL;
	this->columns = NULL;
	reallocNonZeroBuff(nonZeroCountGuess);

	this->rowIndices = new size_vv[this->rowCount + 1];
	INSIST(this->rowIndices != NULL);	

	this->nextRowIndex     = 0;
	this->nextNonZeroIndex = 0;
}

SparseSymMat_CSR::~SparseSymMat_CSR()
{
	delete this->values;
	delete this->columns;
	delete [] this->rowIndices;
}

void 
SparseSymMat_CSR::SetRow(size_v iRow, const precise* rowValues, 
						 const size_v* rowColumns, size_vv rowNonZeroCount)
{
	INSIST(iRow == this->nextRowIndex);
	INSIST(iRow < this->GetRowCount());

	makeSpaceForNonZeros(rowNonZeroCount);

	this->rowIndices[this->nextRowIndex] = this->nextNonZeroIndex + 1;
	for(size_v iElem = 0; iElem < rowNonZeroCount; iElem++)
	{
		INSIST(rowColumns[iElem] >= iRow);
		INSIST(rowColumns[iElem] < this->colCount);
		this->values[this->nextNonZeroIndex]  = rowValues[iElem];
		this->columns[this->nextNonZeroIndex] = rowColumns[iElem] + 1;
		this->nextNonZeroIndex++;
	}

	this->nextRowIndex++;
	if(this->nextRowIndex == this->GetRowCount())
	{
		this->rowIndices[this->nextRowIndex] = this->nextNonZeroIndex + 1;  //one based indexing
		reallocNonZeroBuff(this->nextNonZeroIndex);
	}
}


void 
SparseSymMat_CSR::SetRow(size_v iRow, const vector<precise>& rowValues, 
						 const vector<size_v>& rowColumns)
{	
	INSIST(iRow == this->nextRowIndex);
	INSIST(iRow < this->GetRowCount());
	size_vv rowNonZeroCount = (size_vv) rowValues.size();
	INSIST((size_t) rowNonZeroCount == rowColumns.size());

	makeSpaceForNonZeros(rowNonZeroCount);

	this->rowIndices[this->nextRowIndex] = this->nextNonZeroIndex + 1; //one based indexing
	for(size_v iElem = 0; iElem < rowNonZeroCount; iElem++)
	{
		INSIST(rowColumns[iElem] >= iRow);
		INSIST(rowColumns[iElem] < this->colCount);
		this->values[this->nextNonZeroIndex]  = rowValues[iElem];
		this->columns[this->nextNonZeroIndex] = rowColumns[iElem] + 1; //one based indexing
		this->nextNonZeroIndex++;
	}

	this->nextRowIndex++;
	if(this->nextRowIndex == this->GetRowCount())
	{
		this->rowIndices[this->nextRowIndex] = this->nextNonZeroIndex + 1;  //one based indexing
		reallocNonZeroBuff(this->nextNonZeroIndex);
	}	
}


void 
SparseSymMat_CSR::Multiply(size_v inLen, size_v outLen, 
						   const precise* inVec, precise* outVec) const
{
#ifdef __USE_MKL__
	Multiply_mkl(inLen, outLen, inVec, outVec);
#else
	Multiply_pro(inLen, outLen, inVec, outVec);
#endif
}

void 
SparseSymMat_CSR::makeSpaceForNonZeros(size_vv rowNonZeroCount)
{
	size_vv nonZerosRequired = this->nextNonZeroIndex + rowNonZeroCount;
	if(nonZerosRequired > this->nonZeroCount)
	{
		float currRowFrac = float(this->rowCount) / (this->nextRowIndex + 1);
		int nonZeroCountGuess = (int) (currRowFrac * nonZerosRequired + 100);
		reallocNonZeroBuff(nonZeroCountGuess);
	}
}


void
SparseSymMat_CSR::reallocNonZeroBuff(size_vv nonZeroCountGuess)
{
	this->nonZeroCount = nonZeroCountGuess;

	this->values  = (precise*) realloc(this->values,  this->nonZeroCount * sizeof(*this->values));
	this->columns = (size_v*)  realloc(this->columns, this->nonZeroCount * sizeof(*this->columns));
	INSIST(this->values  != NULL);
	INSIST(this->columns != NULL);
}


