#include "Filter_Driver.h"
#include "Disk.h"
#include "ImageIO.h"

GS_Image Filter_Driver::Filter(string fileName, const Params &params)
{
	GS_ProblemInput::Params probInputParams = GS_ProblemInput::Params::Default();
	switch(params.type)
	{
	case FT_ELE:
	case FT_NPR:
		probInputParams.loadEmData = true;
		break;

	case FT_COLORIZE:
		if(params.gsf_clzParams.improvedSDI)
			probInputParams.loadEmData = true;
		break;

	case FT_MOTION_BLUR:
		probInputParams.loadFlowField = true;
		break;

	default:
		break;
	}

	GS_ProblemInput probInput(fileName, probInputParams, params.gspParams.channelsType);
	GS_Problem prob(probInput, params.gspParams);
	return Filter_Driver::Filter(&prob, params);
}

GS_Image Filter_Driver::Filter(GS_Problem* problem, const Params &params)
{
	GS_Filter* filter = NULL;
	switch(params.type)
	{
	case FT_NULL:
		filter = new GSF_Null(problem);
		break;

	case FT_SHARPEN:
		filter = new GSF_Sharpen(problem, params.gsf_SharpenParams);
		break;

	case FT_ELE:
		filter = new GSF_ELE(problem, params.gsf_EleParams);
		break;

	case FT_TEST:
		filter = new GSF_Test(problem, params.gsf_testParams);
		break;

	case FT_RELIGHT:
		filter = new GSF_Relight(problem, params.gsf_relightParams);
		break;

	case FT_DEBLOCK:
		filter = new GSF_Deblock(problem, params.gsf_deblockParams);
		break;

	case FT_MOTION_BLUR:
		filter = new GSF_MotionBlur(problem, params.gsf_motionBlurParams);
		break;

	case FT_GRAD_CURVE:
		filter = new GSF_GradCurve(problem, params.gsf_gradCurveParams);
		break;

	case FT_NPR:
		filter = new GSF_NPR(problem, params.gsf_nprParams);
		break;

	case FT_TM:
		filter = new GSF_ToneMap(problem, params.gsf_tmParams);
		break;	

	case FT_COLORIZE:
		filter = new GSF_Colorize(problem, params.gsf_clzParams);
		break;

	default:
		REPORT_FAILURE("Unknown filter type - %i\n", params.type);
		break;
	};
	
	if(params.gssParams.levelsToSolve.size() == 0)
		filter->Filter();
	else
		filter->Filter(params.gssParams.levelsToSolve);
	delete filter;

	GS_Solver solver(problem, params.gssParams);
	/*
	 * GS_FourierSolver is much faster but can not handle spatially varying weights
	 * For apps like - deblocking & tonemapping it is fine. For other apps there will be some artifacts
	 */
	//GS_FourierSolver solver(problem, params.gssParams);
	solver.Solve();	

	return problem->GetResultRGB();
}


void Filter_Driver::TestFilter(string fileName, const Params& params)
{
	switch(params.type)
	{
	case FT_NULL:
		GSF_Null::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_SHARPEN:
		GSF_Sharpen::TestFilters(fileName, params.gssParams.outDir);
		break;

	case FT_ELE:
		GSF_ELE::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_TEST:
		GSF_Test::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_RELIGHT:
		GSF_Relight::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_DEBLOCK:
		GSF_Deblock::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_MOTION_BLUR:
		GSF_MotionBlur::TestFilters(fileName, params.gssParams.outDir);		
		break;

	case FT_GRAD_CURVE:
		GSF_GradCurve::TestFilters(fileName, params.gssParams.outDir);
		break;

	case FT_NPR:
		GSF_NPR::TestFilters(fileName, params.gssParams.outDir);
		break;

	case FT_TM:
		GSF_ToneMap::TestFilters(fileName, params.gssParams.outDir);
		break;

	case FT_COLORIZE:
		GSF_Colorize::TestFilters(fileName, params.gssParams.outDir);
		break;

	default:
		REPORT_FAILURE("Unknown filter type - %i\n", params.type);
		break;
	};	
}

