#include "GSF_Sharpen.h"
#include "Disk.h"
#include "ImageIO.h"

GS_Problem::Params GSF_Sharpen::GS_Problem_Params_Default()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount = 1;
	params.channelsType = CT_RGB;
	return params;
}

GS_Problem::Params GSF_Sharpen::GS_Problem_Params_MultiFreq()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount  = 2;
	params.plsParams.sigmaD      = 12.5f;
	params.plsParams.sigmaS      = 0.03f;
	params.plsParams.sigmaD_mult = 1.0f;
	params.plsParams.levelsType  = GS_ProblemLevels::LT_DOG;
	params.channelsType          = CT_RGB;
	return params;
}

GS_Solver::Params GSF_Sharpen::GS_Solver_Params_Default()
{
	GS_Solver::Params params = GS_Solver::Params::Default();
	params.fidelity_data = 3.0E-02;
	params.maxIters = 200;
	params.useGlobalDatacost = true;
	return params;
}

GS_Image GSF_Sharpen::Sharpen(GS_Image inputImg,
							  Params fltParams,
							  GS_Problem::Params probParams,
							  GS_Solver::Params solParams)
{
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Creating GS problem... \n");
	probParams.Dump(paramsSpacing);
	GS_ProblemInput probInput(inputImg, probParams.channelsType);
	GS_Problem prob(probInput, probParams);	

	return GSF_Sharpen::Sharpen(&prob, fltParams, solParams);
}


GS_Image GSF_Sharpen::Sharpen(GS_Problem* prob,
							  Params fltParams,
							  GS_Solver::Params solParams)
{
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Filtering GS problem... \n");
	fltParams.Dump(paramsSpacing);
	GSF_Sharpen sharpenFlt(prob, fltParams);
	if(solParams.levelsToSolve.size() == 0)
		sharpenFlt.Filter();
	else
		sharpenFlt.Filter(solParams.levelsToSolve);

	REPORT("Solving GS problem... \n");
	solParams.Dump(paramsSpacing);
	GS_Solver solver(prob, solParams);
	solver.Solve();

	return prob->GetResultRGB();	
}

GS_Image GSF_Sharpen::SharpenFreq(GS_Image inputImg, int iFreqStart, int iFreqEnd,
								  Params fltParams,
								  GS_Problem::Params probParams,					
								  GS_Solver::Params solParams)
{
	INSIST(iFreqStart <= iFreqEnd);
	INSIST(iFreqStart >= 0);
	INSIST(iFreqEnd < probParams.plsParams.levelCount);

	INSIST(solParams.levelsToSolve.size() == 0);
	for(int iFreq = iFreqStart; iFreq <= iFreqEnd; iFreq++)
	{
		solParams.levelsToSolve.insert(iFreq);
	}

	return Sharpen(inputImg, fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenHiFreq(GS_Image inputImg,
					                 Params fltParams,
					                 GS_Problem::Params probParams,
				                     GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((0.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenMidFreq(GS_Image inputImg,
					                 Params fltParams,
					                 GS_Problem::Params probParams,
				                     GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;

	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenLowFreq(GS_Image inputImg,
								    Params fltParams,
					                GS_Problem::Params probParams,
				                    GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((3.0f / 3.0f) * probParams.plsParams.levelCount);
	INSIST(iFreqEnd > iFreqStart);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}


GS_Image GSF_Sharpen::Sharpen(GS_Image inputImg,
							  Params startFltParams, Params endFltParams,
							  GS_Problem::Params probParams,
							  GS_Solver::Params solParams)
{
	INCR_TAB_LEVEL();
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Creating GS problem... \n");
	probParams.Dump(paramsSpacing);
	GS_ProblemInput probInput(inputImg, probParams.channelsType);
	GS_Problem prob(probInput, probParams);	

	REPORT("Filtering GS problem... \n");
	GSF_Sharpen sharpenFlt(&prob, startFltParams);
	INSIST(solParams.levelsToSolve.size() == 0);
	INSIST(probParams.plsParams.levelCount > 1);
	for(int iLevel = 0; iLevel < probParams.plsParams.levelCount; iLevel++)
	{	
		float levelFrac = ((float) iLevel / (probParams.plsParams.levelCount - 1));
		Params levelFltParams = Params::Interpolate(startFltParams, endFltParams, levelFrac);
		levelFltParams.Dump(paramsSpacing);
		sharpenFlt.FilterLevel(iLevel, &levelFltParams);
	}		

	REPORT("Solving GS problem... \n");
	solParams.Dump(paramsSpacing);
	GS_Solver solver(&prob, solParams);
	solver.Solve();

	DCR_TAB_LEVEL();
	return prob.GetResultRGB();	
}


void GSF_Sharpen::TestFilters(string fileName, string outDir)
{
	GS_Image inputImg;
	ImageIO::ReadFile(inputImg, fileName);

	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(fileName));	
	GS_Image fltImg;

	Params sharpenParams = GSF_Sharpen::Params::Default();
	Params blurParams    = GSF_Sharpen::Params::Blur();
	Params noopParams(1.0f);
	
	//sharpen tests
	fltImg = GSF_Sharpen::Sharpen(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-default.tga");

	fltImg = GSF_Sharpen::SharpenLowFreq(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-LowFreq.tga");

	//fltImg = GSF_Sharpen::Sharpen(inputImg, blurParams, sharpenParams,  
	//							  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	//ImageIO::WriteFile(fltImg, outDir + fileID + "-blur2sharp.tga");
}



/* pre siggraph
GS_Problem::Params GSF_Sharpen::GS_Problem_Params_Default()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount = 1;
	params.channelsType = CT_RGB;
	return params;
}

GS_Problem::Params GSF_Sharpen::GS_Problem_Params_MultiFreq()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount  = 3;
	params.plsParams.sigmaD      = 10.0f;
	params.plsParams.sigmaD_mult = 1.0f;
	params.plsParams.levelsType  = GS_ProblemLevels::LT_DOG;
	params.channelsType          = CT_RGB;
	return params;
}


GS_Solver::Params GSF_Sharpen::GS_Solver_Params_Default()
{
	GS_Solver::Params params = GS_Solver::Params::Default();
	params.fidelity_data = 3.0E-02;
	return params;
}


GS_Image GSF_Sharpen::Sharpen(GS_Image inputImg,
							  Params fltParams,
							  GS_Problem::Params probParams,
							  GS_Solver::Params solParams)
{
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Creating GS problem... \n");
	probParams.Dump(paramsSpacing);
	GS_ProblemInput probInput(inputImg);
	GS_Problem prob(probInput, probParams);	

	return GSF_Sharpen::Sharpen(&prob, fltParams, solParams);
}


GS_Image GSF_Sharpen::Sharpen(GS_Problem* prob,
							  Params fltParams,
							  GS_Solver::Params solParams)
{
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Filtering GS problem... \n");
	fltParams.Dump(paramsSpacing);
	GSF_Sharpen sharpenFlt(prob, fltParams);
	if(solParams.levelsToSolve.size() == 0)
		sharpenFlt.Filter();
	else
		sharpenFlt.Filter(solParams.levelsToSolve);

	REPORT("Solving GS problem... \n");
	solParams.Dump(paramsSpacing);
	GS_Solver solver(prob, solParams);
	solver.Solve();

	return prob->GetResultRGB();	
}

GS_Image GSF_Sharpen::SharpenFreq(GS_Image inputImg, int iFreqStart, int iFreqEnd,
								  Params fltParams,
								  GS_Problem::Params probParams,					
								  GS_Solver::Params solParams)
{
	INSIST(iFreqStart <= iFreqEnd);
	INSIST(iFreqStart >= 0);
	INSIST(iFreqEnd < probParams.plsParams.levelCount);

	INSIST(solParams.levelsToSolve.size() == 0);
	for(int iFreq = iFreqStart; iFreq <= iFreqEnd; iFreq++)
	{
		solParams.levelsToSolve.insert(iFreq);
	}

	return Sharpen(inputImg, fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenHiFreq(GS_Image inputImg,
					                 Params fltParams,
					                 GS_Problem::Params probParams,
				                     GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((0.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenMidFreq(GS_Image inputImg,
					                 Params fltParams,
					                 GS_Problem::Params probParams,
				                     GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;

	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}

GS_Image GSF_Sharpen::SharpenLowFreq(GS_Image inputImg,
								    Params fltParams,
					                GS_Problem::Params probParams,
				                    GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((3.0f / 3.0f) * probParams.plsParams.levelCount);
	INSIST(iFreqEnd > iFreqStart);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return SharpenFreq(inputImg, iFreqStart, iFreqEnd,
					   fltParams, probParams, solParams);
}


GS_Image GSF_Sharpen::Sharpen(GS_Image inputImg,
							  Params startFltParams, Params endFltParams,
							  GS_Problem::Params probParams,
							  GS_Solver::Params solParams)
{
	INCR_TAB_LEVEL();
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Creating GS problem... \n");
	probParams.Dump(paramsSpacing);
	GS_ProblemInput probInput(inputImg, probParams.channelsType);
	GS_Problem prob(probInput, probParams);	

	REPORT("Filtering GS problem... \n");
	GSF_Sharpen sharpenFlt(&prob, startFltParams);
	INSIST(solParams.levelsToSolve.size() == 0);
	INSIST(probParams.plsParams.levelCount > 1);
	for(int iLevel = 0; iLevel < probParams.plsParams.levelCount; iLevel++)
	{	
		float levelFrac = ((float) iLevel / (probParams.plsParams.levelCount - 1));
		Params levelFltParams = Params::Interpolate(startFltParams, endFltParams, levelFrac);
		levelFltParams.Dump(paramsSpacing);
		sharpenFlt.FilterLevel(iLevel, &levelFltParams);
	}		

	REPORT("Solving GS problem... \n");
	solParams.Dump(paramsSpacing);
	GS_Solver solver(&prob, solParams);
	solver.Solve();

	DCR_TAB_LEVEL();
	return prob.GetResultRGB();	
}


void GSF_Sharpen::TestFilters(string fileName, string outDir)
{
	GS_Image inputImg;
	ImageIO::ReadFile(inputImg, fileName);

	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(fileName));	
	GS_Image fltImg;

	Params sharpenParams = GSF_Sharpen::Params::Default();
	Params blurParams    = GSF_Sharpen::Params::Blur();
	Params noopParams(1.0f);
	
	//sharpen tests
	fltImg = GSF_Sharpen::Sharpen(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-default.tga");

	fltImg = GSF_Sharpen::SharpenHiFreq(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-HiFreq.tga");

	fltImg = GSF_Sharpen::SharpenMidFreq(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-MidFreq.tga");

	fltImg = GSF_Sharpen::SharpenLowFreq(inputImg, sharpenParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-LowFreq.tga");

	fltImg = GSF_Sharpen::Sharpen(inputImg, noopParams, sharpenParams, 
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-hi2low.tga");

	fltImg = GSF_Sharpen::Sharpen(inputImg, sharpenParams, noopParams,  
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharpen-low2hi.tga");

	//blur tests
	fltImg = GSF_Sharpen::Sharpen(inputImg, blurParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-default.tga");

	fltImg = GSF_Sharpen::SharpenHiFreq(inputImg, blurParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-HiFreq.tga");

	fltImg = GSF_Sharpen::SharpenMidFreq(inputImg, blurParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-MidFreq.tga");

	fltImg = GSF_Sharpen::SharpenLowFreq(inputImg, blurParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-LowFreq.tga");

	fltImg = GSF_Sharpen::Sharpen(inputImg, noopParams, blurParams, 
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-hi2low.tga");

	fltImg = GSF_Sharpen::Sharpen(inputImg, blurParams, noopParams,  
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur-low2hi.tga");

	///mixed tests
	fltImg = GSF_Sharpen::Sharpen(inputImg, sharpenParams, blurParams, 
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-sharp2blur.tga");

	fltImg = GSF_Sharpen::Sharpen(inputImg, blurParams, sharpenParams,  
								  GS_Problem_Params_MultiFreq(), GS_Solver_Params_Default());
	ImageIO::WriteFile(fltImg, outDir + fileID + "-blur2sharp.tga");
}

*/

