#include "GSF_ToneMap.h"
#include "ImageProcessing.h"
#include "ImageIO.h"

GSF_ToneMap::GSF_ToneMap(GS_Problem* _problem, const Params &_params) : GS_Filter(_problem)
{
	this->params = _params;
}

void GSF_ToneMap::FilterLevel(int iLevel, void* userData)
{	
	INSIST(iLevel == 0);

	Params* levelParams = (userData == NULL) ? &this->params : (Params*) userData;

	GS_ProblemLevel* levelProb = this->problem->GetLevel(iLevel);

	GS_Image levelOrg,  levelFlt;
	GS_Image levelData, levelDataW;
	GS_Image levelOrgGF_X, levelOrgGF_Y;
	GS_Image levelFltGF_X, levelFltGF_Y;
	levelProb->GetData(levelOrg, levelFlt,
					   levelData, levelDataW,
					   levelOrgGF_X, levelOrgGF_Y, levelFltGF_X, levelFltGF_Y);

	CShape imgShape = levelOrg.Shape();
	GS_Image attentuationImg(imgShape);
	getAttentuationImg(levelOrg, 0, levelParams, attentuationImg);

	uint pixelCount = imgShape.PixelCount();
	for(uint iPixel = 0; iPixel < pixelCount; iPixel++)
	{
		float attenWeight = attentuationImg[iPixel];
		INSIST(attenWeight >= 0.0f);
		levelFltGF_X[iPixel] = attenWeight * levelOrgGF_X[iPixel];
		levelFltGF_Y[iPixel] = attenWeight * levelOrgGF_Y[iPixel];
	}	

	uint iPixel = 0;
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++)
	for(int channel = 0; channel < imgShape.nBands;  channel++, iPixel++)
	{
		float attenWeight = attentuationImg.Pixel(x, y, channel);
		INSIST(attenWeight >= 0.0f);
		levelFltGF_X[iPixel] = attenWeight * levelOrgGF_X[iPixel];
		levelFltGF_Y[iPixel] = attenWeight * levelOrgGF_Y[iPixel];		
	}	
}

void GSF_ToneMap::getAttentuationImg(const GS_Image& img, int iLevel, const Params* levelParams, GS_Image& attenImg, float avgGradMagGlobal)
{
	const int MIN_CHILD_DIM = this->params.minLevelSize * 2;
	INSIST(MIN_CHILD_DIM >= 8);
	const static float PYRAMID_SAMPLE_SCALE = 2.0f;
	
	CShape imgShape = img.Shape();
	INSIST(attenImg.Shape() == imgShape);

	GS_Image imgGradMag(imgShape);
	float avgGradMag = genGradData(img, iLevel, imgGradMag);
	if(iLevel == 0)
		avgGradMagGlobal = avgGradMag;
	attenImg = getWeightImg(imgGradMag, levelParams, avgGradMagGlobal);

	if((imgShape.width >= MIN_CHILD_DIM) && (imgShape.height >= MIN_CHILD_DIM))
	{
		GS_Image imgNextLevel;
		ImageProcessing::ScaleImage(imgNextLevel, img, 1 / PYRAMID_SAMPLE_SCALE, true, 3.0f);

		GS_Image attenImgNextLevel(imgNextLevel.Shape());
		getAttentuationImg(imgNextLevel, iLevel + 1, levelParams, attenImgNextLevel, avgGradMagGlobal);

		GS_Image attenImgUpSampled;
		ImageProcessing::ScaleImage(attenImgUpSampled, attenImgNextLevel, PYRAMID_SAMPLE_SCALE, false);

		CShape upSampleShape = attenImgUpSampled.Shape();
		uint pixelAddr = 0;
		for(int y = 0; y < imgShape.height; y++)
		for(int x = 0; x < imgShape.width;  x++)				
		for(int iChannel = 0; iChannel < imgShape.nBands; iChannel++, pixelAddr++)
		{
			float upSampledWeight = attenImgUpSampled.Pixel(min(x, upSampleShape.width - 1), 
														    min(y, upSampleShape.height - 1), 
														    iChannel);
			SET_IF_GT(upSampledWeight, 0.0f);
			SET_IF_LT(upSampledWeight, 1.0f);
			attenImg[pixelAddr] *= upSampledWeight;
		}
	}
}

float GSF_ToneMap::genGradData(const GS_Image& img, int iLevel, GS_Image& imgGradMag)
{
	CShape imgShape = img.Shape();
	INSIST(imgGradMag.Shape() == imgShape);
	
	INSIST(iLevel >= 0);
	//check
	float gradDenom = pow(1.0f, iLevel + 1);

	float gradMagSum    = 0.0f;
	int pixelAddr       = 0;
	int rightPixelAddr  = imgShape.nBands;
	int leftPixelAddr   = -rightPixelAddr;	
	int topPixelAddr    = (imgShape.nBands * imgShape.width);
	int bottomPixelAddr = -topPixelAddr;	
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++)
	for(int iChannel = 0; iChannel < imgShape.nBands; iChannel++, 
		pixelAddr++, leftPixelAddr++, rightPixelAddr++, bottomPixelAddr++, topPixelAddr++)
	{
		float gx, gy;

		if((x > 0) && (x < imgShape.width - 1))
			gx = (img[rightPixelAddr] - img[leftPixelAddr]) / gradDenom;
		else
			gx = 0;

		if((y > 0) && (y < imgShape.height - 1))
			gy = (img[topPixelAddr] - img[bottomPixelAddr]) / gradDenom;
		else
			gy = 0;

		imgGradMag[pixelAddr] = sqrt((gx * gx) + (gy * gy));
		gradMagSum += imgGradMag[pixelAddr];
	}	

	INSIST(gradMagSum > 0.0f);
	uint pixelCount = imgShape.PixelCount();
	INSIST(pixelCount > 0);	
	return (gradMagSum / pixelCount);
}

GS_Image GSF_ToneMap::getWeightImg(const GS_Image& imgGradMag, const Params* levelParams, float avgGradMag)
{
	CShape imgShape = imgGradMag.Shape();
	GS_Image weightImg(imgShape);

	float gamma = levelParams->attenFuncScale * avgGradMag;

	uint pixelCount = imgShape.PixelCount();
	INSIST(pixelCount > 0);
	for(uint iPixel = 0; iPixel < pixelCount; iPixel++)
	{
		float gradMag = imgGradMag[iPixel];
		INSIST(gradMag >= 0.0f);

		if(gradMag > gamma)
		{
			float weightDenom = gradMag / gamma;
			float weightNum   = pow(weightDenom, levelParams->attenFuncExp);
			float weight      = weightNum / weightDenom;
			weightImg[iPixel] = weight;
		}
		else
		{
			weightImg[iPixel] = 1.0f;
		}
	}

	return weightImg;
}

