#include "GF_Integrator_Laspack.h"


GF_Integrator_Laspack::GF_Integrator_Laspack(CFloatImage _orgImg, 
											 CFloatImage _solImgGx,
											 CFloatImage _solImgGy,
											 GF_Integrator_Laspack::GF_Integrator_Params _params)
{
	this->orgImg   = _orgImg;
	this->solImgGx = _solImgGx;
	this->solImgGy = _solImgGy;
	this->params   = _params;
}

void GF_Integrator_Laspack::Solve(CFloatImage &solImg)
{
	CShape imgShape = solImg.Shape(); 
	INSIST(imgShape == this->orgImg.Shape());
	INSIST(imgShape == this->solImgGx.Shape());
	INSIST(imgShape == this->solImgGy.Shape());

	for(int iChannel = 0; iChannel < imgShape.nBands; iChannel++)
	{
		REPORT("Integrating channel - %i\n", iChannel);
		integrateChannel(iChannel, solImg);	
	}
}

void GF_Integrator_Laspack::integrateChannel(int iChannel, CFloatImage &solImg)
{
	CShape imgShape = solImg.Shape();

	int numOfVars = imgShape.width * imgShape.height;
	INSIST(numOfVars > 0);
	SparseSolver solver(numOfVars, true);
	defineChannelEqs(iChannel, solver);
	solver.Solve(this->params.accuracy, this->params.maxIters);
		
	uint pixelAddr = 0;
	uint nodeAddr  = 0;
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++, pixelAddr += imgShape.nBands, nodeAddr++)
	{
		solImg[pixelAddr + iChannel] = (float) solver.GetSolution(nodeAddr);
	}
}

void GF_Integrator_Laspack::defineChannelEqs(int iChannel, SparseSolver &solver)
{
	CShape imgShape = this->orgImg.Shape();
	ENSURE((iChannel >= 0) && (iChannel < imgShape.nBands));

	const int EquationVars     = 5;
	const int dx[EquationVars] = { 0, -1,  0,  1,  0};
	const int dy[EquationVars] = {-1,  0,  0,  0,  1};

	int    eqVarAddr[EquationVars];
	double eqVarCoef[EquationVars];

	uint nodeAddr  = 0;
	uint pixelAddr = 0;
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++, pixelAddr += imgShape.nBands, nodeAddr++)
	{
		double rhs = 0;
		double fxy = this->orgImg[pixelAddr + iChannel];

		int   eqVarCount   = 0;
		int   nodeVarIndex = -1;
		float nodeVarCoef  = 0.0f;
		for(int iEqVar = 0; iEqVar < EquationVars; iEqVar++)
		{
			if(iEqVar == 2)
			{
				ENSURE(dx[iEqVar] == 0);
				ENSURE(dy[iEqVar] == 0);
				eqVarAddr[eqVarCount] = nodeAddr;
				nodeVarIndex = eqVarCount;
				eqVarCount++;

				rhs         += this->params.fidelity * fxy;
				nodeVarCoef += this->params.fidelity;

				continue;
			}

			int neighbourX = x + dx[iEqVar];
			int neighbourY = y + dy[iEqVar];
			if(imgShape.InBounds(neighbourX, neighbourY) == false) continue;
	
			uint neighbourNodeAddr  = imgShape.NodeIndex(neighbourX, neighbourY);
			uint neighbourPixelAddr = neighbourNodeAddr * imgShape.nBands;
			float neighCoef;
			float neighGrad;
			if(neighbourY != y)
			{
				INSIST(neighbourX == x);
				if(y > neighbourY)
				{
					neighCoef = 1.0f;
					neighGrad = this->solImgGy[pixelAddr + iChannel];
				}
				else
				{
					neighCoef = 1.0f;
					neighGrad = -this->solImgGy[neighbourPixelAddr + iChannel];
				}				
			}
			else
			{
				INSIST(neighbourX != x);
				if(x > neighbourX)
				{
					neighCoef = 1.0f;
					neighGrad = this->solImgGx[pixelAddr + iChannel];
				}
				else
				{
					neighCoef = 1.0f;
					neighGrad = -this->solImgGx[neighbourPixelAddr + iChannel];
				}
			}

			if(neighCoef != 0.0f)
			{
				if(neighbourNodeAddr >= nodeAddr)
				{
					eqVarAddr[eqVarCount] = neighbourNodeAddr;
					eqVarCoef[eqVarCount] = -neighCoef;
					eqVarCount++;
				}

				rhs += neighCoef * neighGrad;
				nodeVarCoef += neighCoef;
			}
		}

		eqVarCoef[nodeVarIndex] = nodeVarCoef;
		ENSURE(nodeVarCoef > 0.0);
		solver.SetRow(nodeAddr, eqVarCount, fxy, rhs, eqVarAddr, eqVarCoef);
	}
}