#include "LinearSolver.h"
#include "SparseSqrSymMat.h"



//Solve linear system using conjugate gradients
//Algorithm follows the B2 outline found in www.cs.cmu.edu/~quake-papers/painless-conjugate-gradient.pdf 
void LinearSolver::Solve_CG(int n, AbstractMat *A, precise *x, precise *b, 
							LinearSolverParams params)
{
	INSIST(A->GetRowCount() == n);
	INSIST(A->GetColCount() == n);
	
	precise *r    = new precise[n]; //r = residual vector	
	precise *d    = new precise[n]; //d = conjugate direction	
	precise *Av   = new precise[n]; //Av holds the vec result from A * v (where v = r, x, or d)	
	precise rtr;                    //rtr = r' * r	
	precise rtrOrg;                 //rtrOrg = rtr at the first iteration
	int iter;                       //iterations
	precise epsilonSqr;             //error tolerance squared            
	precise rtrMax;                 //the maximum allowable residual
	precise alpha;                  //length to travel in the direction d 

	INSIST(r  != NULL);
	INSIST(Av != NULL);

	//r = b - Ax
	A->Multiply(n, n, x, Av);   // Av  = A * x
	for(int i = 0; i < n; i++)
	{
		r[i] = b[i] - Av[i];
	}

	//d = r
	Vec::Copy(n, d, r);
	
	iter       = 0;
	rtr        = Vec::Multiply(n, r, r);
	rtrOrg     = rtr;	
	
	epsilonSqr = params.errTolerance * params.errTolerance;
	//rtrMax     = rtrOrg * epsilonSqr;
	rtrMax     = epsilonSqr;
	 
	while((iter < params.maxIters) &&
		  (rtr > rtrMax))
	{
		A->Multiply(n, n, d, Av);                      //Av = A * d
		precise alphaDenom = Vec::Multiply(n, d, Av);  //alphaDenom = d' * (A * d)

		if(alphaDenom != 0.0f)
			alpha = rtr / alphaDenom;  //alpha = rtr / (d' * (A * d))
		else // d is the null vector
		{			
			break;
		}

		//x = x + ad
		for(int i = 0; i < n; i++)
		{
			x[i] += alpha * d[i];
		}

		if((iter % params.fltErrMaxiters == 0) &&
		   (iter > 0)) 
		{
			//r = b - Ax
			A->Multiply(n, n, x, Av);  
			for(int i = 0; i < n; i++)
			{
				r[i] = b[i] - Av[i];
			}
		}
		else
		{
			//r = r - (alpha * (A * r))
			for(int i = 0; i < n; i++)
			{
				r[i] -= alpha * Av[i];
			}
		}


		precise rtrOld = rtr;
		rtr = Vec::Multiply(n, r, r);
		precise beta = rtr / rtrOld;

		//d = r + beta*d
		for(int i = 0; i < n; i++)
		{
			d[i] = r[i] + (beta * d[i]);
		}

		iter++;
	}

	PRINT_SCI(rtrOrg);
	PRINT_SCI(rtr);
	PRINT_INT(iter);

	delete [] r;
	delete [] d;
	delete [] Av;
}

precise LinearSolver::GetResidual(int n, AbstractMat *A, precise *x, precise *b)
{
	precise *Ax = new precise[n]; //Av holds the vec result from A * x

	//r = b - Ax
	precise residual = 0.0;
	A->Multiply(n, n, x, Ax);  
	for(int i = 0; i < n; i++)
	{
		precise err = b[i] - Ax[i];
		residual += err * err;
	}

	delete [] Ax;

	return residual;
}



//Solve linear system using gradient descent
//Algorithm follows the B1 outline found in www.cs.cmu.edu/~quake-papers/painless-conjugate-gradient.pdf 
//Bug - Does not minimize the energy!
void LinearSolver::Solve_GD(int n, AbstractMat *A, precise *x, precise *b, LinearSolverParams params)
{
	INSIST(A->GetRowCount() == n);
	INSIST(A->GetColCount() == n);
	
	precise *r    = new precise[n]; //r = residual vector	
	precise *Av   = new precise[n]; //Av holds the vec result from A * v (where v = r or x)	
	precise rtr;                    //rtr = r' * r	
	precise rtrOrg;                 //rtrOrg = rtr at the first iteration
	int iter;                       //iterations
	precise epsilonSqr;             //error tolerance squared            
	precise alpha;                  //length to travel in the direction r 

	INSIST(r  != NULL);
	INSIST(Av != NULL);

	//r = b - Ax
	A->Multiply(n, n, x, Av);   // Av  = A * x
	for(int i = 0; i < n; i++)
	{
		r[i] = b[i] - Av[i];
	}
	
	rtr        = Vec::Multiply(n, r, r);
	rtrOrg     = rtr;
	iter       = 0;
	epsilonSqr = params.errTolerance * params.errTolerance;
	 
	//bug
	//while((iter < params.maxIters) &&
	//	  (rtr > epsilonSqr * rtrOrg))

	while((iter < params.maxIters) && (rtr > 0))
	{
		A->Multiply(n, n, r, Av);               // Av = A * x
		alpha = rtr / Vec::Multiply(n, r, Av);  //alpha = rtr / (r' * (A * r))

		//x = x + ar
		for(int i = 0; i < n; i++)
		{
			x[i] += alpha * r[i];
		}


		//if(((iter % params.fltErrMaxiters) == 0) &&
		//   (iter > 0))
		//{
			//r = b - Ax
			A->Multiply(n, n, x, Av);   // Av  = A * x
			for(int i = 0; i < n; i++)
			{
				r[i] = b[i] - Av[i];
			}
		//}
		//else
		//{
		//	//r = r - (alpha * (A * r))
		//	for(int i = 0; i < n; i++)
		//	{
		//		r[i] -= alpha * Av[i];
		//	}
		//}

		rtr = Vec::Multiply(n, r, r);
		iter++;

		PRINT_FLOAT(rtr);
	}

	//bug
	//exit(0);

	delete [] r;
	delete [] Av;
}

//Solve linear system using gradient descent
//Algorithm follows the B3 outline found in www.cs.cmu.edu/~quake-papers/painless-conjugate-gradient.pdf 
void LinearSolver::Solve_PCG(int n, AbstractMat *A, precise *x, precise *b, LinearSolverParams params,
							 AbsractPreconditioner *MInv)
{

	NoopPreconditioner test;

	INSIST(A->GetRowCount() == n);
	INSIST(A->GetColCount() == n);
	
	precise *r    = new precise[n]; //r = residual vector	
	precise *d    = new precise[n]; //d = conjugate direction	
	precise *s    = new precise[n]; //s = preconditioned r
	precise *Av   = new precise[n]; //Av holds the vec result from A * v (where v = r, x, or d)	
	precise rtr;                    //rtr = r' * r	
	precise rtrOrg;                 //rtrOrg = rtr at the first iteration
	int iter;                       //iterations
	precise epsilonSqr;             //error tolerance squared            
	precise rtrMax;                 //the maximum allowable residual
	precise alpha;                  //length to travel in the direction d 

	INSIST(r  != NULL);
	INSIST(Av != NULL);

	//r = b - Ax
	A->Multiply(n, n, x, Av);   // Av  = A * x
	for(int i = 0; i < n; i++)
	{
		r[i] = b[i] - Av[i];
	}

	//d = MInv * r
	Vec::Copy(n, d, r);
	MInv->Apply(n, d);
	
	iter       = 0;
	rtr        = Vec::Multiply(n, r, d);
	rtrOrg     = rtr;	
	epsilonSqr = params.errTolerance * params.errTolerance;
	//rtrMax     = rtrOrg * epsilonSqr;
	rtrMax     = epsilonSqr;

	while((iter < params.maxIters) &&
		  (rtr > rtrMax))
	{
		A->Multiply(n, n, d, Av);                      //Av = A * d
		precise alphaDenom = Vec::Multiply(n, d, Av);  //alphaDenom = d' * (A * d)

		if(alphaDenom != 0.0f)
			alpha = rtr / alphaDenom;  //alpha = rtr / (d' * (A * d))
		else // d is the null vector
		{			
			break;
		}

		//x = x + ad
		for(int i = 0; i < n; i++)
		{
			x[i] += alpha * d[i];
		}

		if((iter % params.fltErrMaxiters == 0) &&
		   (iter > 0)) 
		{
			//r = b - Ax
			A->Multiply(n, n, x, Av);  
			for(int i = 0; i < n; i++)
			{
				r[i] = b[i] - Av[i];
			}
		}
		else
		{
			//r = r - (alpha * (A * r))
			for(int i = 0; i < n; i++)
			{
				r[i] -= alpha * Av[i];
			}
		}

		//s = Minv * r
		Vec::Copy(n, s, r);
		MInv->Apply(n, s);

		precise rtrOld = rtr;
		rtr = Vec::Multiply(n, r, s);
		precise beta = rtr / rtrOld;

		//d = s + beta*d
		for(int i = 0; i < n; i++)
		{
			d[i] = s[i] + (beta * d[i]);
		}

		iter++;
	}

	PRINT_SCI(rtrOrg);
	PRINT_SCI(rtr);
	PRINT_INT(iter);

	delete [] r;
	delete [] d;
	delete [] s;
	delete [] Av;
}

