#include "ImageIO.h"
#include "ImageProcessing.h"
#include "jpeglib.h"


void 
ImageIO::WriteFileJPG(const CByteImage& img, const char* filename, 
					  const CompressionInfo& compressionInfo)
{
	FILE * outfile;    
    if ((outfile = fopen(filename, "wb")) == NULL) 
	{		
        REPORT_FAILURE("Could not open jpg file - %s\n", filename);
    }
    
	struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
	j_compress_ptr cinfoPtr = &cinfo;
    cinfo.err = jpeg_std_error(&jerr);

    jpeg_create_compress(cinfoPtr);
	jpeg_stdio_dest(cinfoPtr, outfile);

	CShape imgShape        = img.Shape();
	cinfo.image_width      = imgShape.width;
	cinfo.image_height     = imgShape.height;
	cinfo.input_components = imgShape.nBands;
	cinfo.in_color_space   = JCS_RGB;
	INSIST((imgShape.nBands == 1) || (imgShape.nBands == 3));

	jpeg_set_defaults(cinfoPtr);

	CHECK_RANGE(compressionInfo.quality, 1, 100);
	jpeg_set_quality(cinfoPtr, compressionInfo.quality, TRUE);

	cinfo.comp_info[0].h_samp_factor = compressionInfo.h_samp_factor[0];
	cinfo.comp_info[0].v_samp_factor = compressionInfo.v_samp_factor[0];
	if(imgShape.nBands == 3)
	{
		cinfo.comp_info[1].h_samp_factor = compressionInfo.h_samp_factor[1];
		cinfo.comp_info[1].v_samp_factor = compressionInfo.v_samp_factor[1];
		cinfo.comp_info[2].h_samp_factor = compressionInfo.h_samp_factor[2];
		cinfo.comp_info[2].v_samp_factor = compressionInfo.v_samp_factor[2];
	}
	else
	{
		cinfo.comp_info[1].h_samp_factor = cinfo.comp_info[1].v_samp_factor = 0;
		cinfo.comp_info[2].h_samp_factor = cinfo.comp_info[2].v_samp_factor = 0;
	}

	jpeg_start_compress(cinfoPtr, TRUE);
	
	JSAMPROW row_pointer[1];    
	while (cinfo.next_scanline < cinfo.image_height) 
	{
		row_pointer[0] = (JSAMPROW) img.PixelAddress(0,  imgShape.height - 1 - cinfo.next_scanline, 0);
		jpeg_write_scanlines(cinfoPtr, row_pointer, 1);
	}

	jpeg_finish_compress(cinfoPtr);
	jpeg_destroy_compress(cinfoPtr);

	fclose(outfile);
}

void 
ImageIO::WriteFileJPG(const CFloatImage& img, const char* filename, 
					  const CompressionInfo& compressionInfo)
{
	CByteImage byteImage = ImageProcessing::FloatToByteImg(img);
	ImageIO::WriteFileJPG(byteImage, filename, compressionInfo);
}

void 
ImageIO::WriteFileJPG(const CHalfImage& img, const char* filename,
				 	  const CompressionInfo& compressionInfo)
{
	CByteImage byteImage = ImageProcessing::HalfToByteImg(img);
	ImageIO::WriteFileJPG(byteImage, filename, compressionInfo);
}

void 
ImageIO::ReadFileJPG(CByteImage& img, const char* filename, 
					 CompressionInfo* compressionInfo)
{
	CompressionInfo dummyCompressionInfo;
	if(compressionInfo == NULL) compressionInfo = &dummyCompressionInfo;

	FILE * infile;
    if ((infile = fopen(filename, "rb")) == NULL) 
	{
      REPORT_FAILURE("Could not open jpg file - %s\n", filename);
    }    

	struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
	j_decompress_ptr cinfoPtr = &cinfo;
    cinfo.err = jpeg_std_error(&jerr);
	
    jpeg_create_decompress(cinfoPtr);
	jpeg_stdio_src(cinfoPtr, infile);
	jpeg_read_header(cinfoPtr, FALSE);	

	cinfo.do_fancy_upsampling = false;
	jpeg_start_decompress(cinfoPtr);

	CShape imgShape(cinfo.image_width, cinfo.image_height, cinfo.num_components);
	INSIST((imgShape.nBands == 1) || (imgShape.nBands == 3));
	img.ReAllocate(imgShape);

	compressionInfo->h_samp_factor[0] = cinfo.comp_info[0].h_samp_factor;
	compressionInfo->v_samp_factor[0] = cinfo.comp_info[0].v_samp_factor;
	if(imgShape.nBands == 3)
	{
		compressionInfo->h_samp_factor[1] = cinfo.comp_info[1].h_samp_factor;
		compressionInfo->v_samp_factor[1] = cinfo.comp_info[1].v_samp_factor;
		compressionInfo->h_samp_factor[2] = cinfo.comp_info[2].h_samp_factor;
		compressionInfo->v_samp_factor[2] = cinfo.comp_info[2].v_samp_factor;
	}
	else
	{
		compressionInfo->h_samp_factor[1] = compressionInfo->v_samp_factor[1] = 0;
		compressionInfo->h_samp_factor[2] = compressionInfo->v_samp_factor[2] = 0;
	}

	JSAMPROW row_pointer[1];
    while (cinfo.output_scanline < cinfo.image_height) 
	{
       row_pointer[0] = (JSAMPROW) img.PixelAddress(0, imgShape.height - 1 - cinfo.output_scanline, 0);
       jpeg_read_scanlines(cinfoPtr, row_pointer, 1);
    }

	jpeg_finish_decompress(cinfoPtr);
	jpeg_destroy_decompress(cinfoPtr);

	fclose(infile);
}

void 
ImageIO::ReadFileJPG(CFloatImage& img, const char* filename, 
					 CompressionInfo* compressionInfo)
{
	CByteImage byteImage;
	ImageIO::ReadFileJPG(byteImage, filename, compressionInfo);
	img = ImageProcessing::ByteToFloatImg(byteImage);
}


CShape ImageIO::GetImageShapeJpg(const char *filename)
{
	CShape imgShape;

	FILE * infile;
    if ((infile = fopen(filename, "rb")) == NULL) 
	{
      REPORT_FAILURE("Could not open jpg file - %s\n", filename);
    }    

	struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
	j_decompress_ptr cinfoPtr = &cinfo;
    cinfo.err = jpeg_std_error(&jerr);
	
    jpeg_create_decompress(cinfoPtr);
	jpeg_stdio_src(cinfoPtr, infile);
	jpeg_read_header(cinfoPtr, FALSE);	

	imgShape = CShape(cinfo.image_width, cinfo.image_height, cinfo.num_components);
	INSIST((imgShape.nBands == 1) || (imgShape.nBands == 3));

	jpeg_destroy_decompress(cinfoPtr);
	fclose(infile);

	return imgShape;
}

