///////////////////////////////////////////////////////////////////////////
//
// NAME
//  FileIO.cpp -- image file input/output
//
// DESCRIPTION
//  Read/write image files, potentially using an interface to an
//  external package.
//
//  Currently, only Targa input/output is supported (grayscale and color
//  byte images), and only a subset of Targa formats in supported.
//
// SEE ALSO
//  FileIO.h            longer description
//
// Copyright  Richard Szeliski, 2001.  See Copyright.h for more details
//
///////////////////////////////////////////////////////////////////////////

#include "ImageIO.h"
#include "Debug.h"
#include "ImageProcessing.h"

extern const char *IMAGE_IO_EXT_STRS[4] = {"jpg,jpeg",     "tif",                 "tga",              "jpg,jpeg,tif,tga"};
extern const char *IMAGE_IO_EXT_TYPE[4] = {"Jpeg files",   "Tagged image format", "Truevision Targa", "Image files"};


void ImageIO::ReadFile(CFloatImage &img, const char* filename, float scale)
{
	if(scale == 1.0f)
	{
		ReadFile(img, filename);
	}
	else
	{
		CFloatImage tempImg;
		ReadFile(tempImg, filename);
		if(scale > 1.0f)
			ImageProcessing::ScaleImage(img, tempImg, scale, false);
		else
			ImageProcessing::ScaleImage(img, tempImg, scale);
	}
}


void ImageIO::ReadFile(CByteImage& img, const char* filename, float scale)
{
	if(scale == 1.0f)
	{
		ReadFile(img, filename);
	}
	else
	{
		CFloatImage fltImg;
		ReadFile(fltImg, filename, scale);
		img = ImageProcessing::FloatToByteImg(fltImg);
	}
}


/*
void ImageIO::WriteFile(const CFloatImage& img, const char* filename)
{
    // Determine the file extension

	const char *dot = strrchr(filename, '.');
	if(dot == NULL)
	{
		REPORT_FAILURE("No file extention found - %s", filename);
	}

	string ext = (dot + 1);
	ext = tolower(ext);

	if(ext.compare("tga") == 0)
    {
		WriteFileTGA(img, filename);
    }
	else if(ext.compare("tif") == 0)
    {
		WriteFileTiff(img, filename);
	}
	else if((ext.compare("jpg")  == 0) ||
			(ext.compare("jpeg") == 0))
    {
		WriteFileJPG(img, filename);
	}
	else if((ext.compare("hdr")  == 0) ||
			(ext.compare("pic") == 0))
    {
		WriteFileRGBE(img, filename);
	}
	else if(ext.compare("raw") == 0)
	{
		WriteRaw(img, filename);
	}
    else
	{
		REPORT_FAILURE("File type not supported - %s", filename);
	}
}

void ImageIO::WriteFile(const CByteImage &img, const char* filename)
{
    // Determine the file extension
    const char *dot = strrchr(filename, '.');
	if(dot == NULL)
	{
		REPORT_FAILURE("No file extention found - %s", filename);
	}

	string ext = (dot + 1);
	ext = tolower(ext);

	if(ext.compare("tga") == 0)
    {
		WriteFileTGA(img, filename);
    }
	else if(ext.compare("tif") == 0)
    {
		WriteFileTiff(img, filename);
	}
	else if((ext.compare("jpg")  == 0) ||
			(ext.compare("jpeg") == 0))
    {
		WriteFileJPG(img, filename);
	}
	else if((ext.compare("hdr")  == 0) ||
			(ext.compare("pic") == 0))
    {
		WriteFileRGBE(img, filename);
	}
	else if(ext.compare("raw") == 0)
	{
		WriteRaw(img, filename);
	}
    else
	{
		REPORT_FAILURE("File type not supported - %s", filename);
	}
}
*/

void ImageIO::ReadFile(CFloatImage &img, const char* filename)
{
    // Determine the file extension
    const char *dot = strrchr(filename, '.');
	if(dot == NULL)
	{
		REPORT_FAILURE("No file extention found - %s", filename);
	}

	string ext = (dot + 1);
	ext = tolower(ext);

	img.channelsType = CT_RGB;

	if(ext.compare("tga") == 0)
    {
		ReadFileTGA(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
    }
	else if(ext.compare("tif") == 0)
    {
		ReadFileTiff(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
	}
	else if((ext.compare("jpg")  == 0) ||
			(ext.compare("jpeg") == 0))
    {
		ReadFileJPG(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
	}
	else if((ext.compare("hdr")  == 0) ||
			(ext.compare("pic") == 0))
    {
		ReadFileRGBE(img, filename);
		INSIST(img.Shape().nBands == 3);
	}
	else if(ext.compare("raw") == 0)
	{
		ReadRaw(img, filename);
	}
    else
	{
		REPORT_FAILURE("File type not supported - %s", filename);
	}
}

void ImageIO::ReadFile(CByteImage &img, const char* filename)
{
    // Determine the file extension
    const char *dot = strrchr(filename, '.');
	if(dot == NULL)
	{
		REPORT_FAILURE("No file extention found - %s", filename);
	}

	string ext = (dot + 1);
	ext = tolower(ext);

	img.channelsType = CT_RGB;

	if(ext.compare("tga") == 0)
    {
		ReadFileTGA(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
    }
	else if(ext.compare("tif") == 0)
    {
		ReadFileTiff(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
	}
	else if((ext.compare("jpg")  == 0) ||
			(ext.compare("jpeg") == 0))
    {
		ReadFileJPG(img, filename);
		if(img.Shape().nBands == 1)
			img.channelsType = CT_LUM;
	}
	else if((ext.compare("hdr")  == 0) ||
			(ext.compare("pic") == 0))
    {
		ReadFileRGBE(img, filename);
		INSIST(img.Shape().nBands == 3);
	}
	else if(ext.compare("raw") == 0)
	{
		ReadRaw(img, filename);
	}
    else
	{
		REPORT_FAILURE("File type not supported - %s", filename);
	}
}

CShape ImageIO::GetImageShape(const char *fileName, float scale)
{
	CShape imgShape;

	const char *dot = strrchr(fileName, '.');
	if(dot == NULL)
	{
		REPORT_FAILURE("No file extention found - %s", fileName);
	}

	if(strcmp(dot, ".tif") == 0)
    {
		imgShape = GetImageShapeTiff(fileName);
    }
	else if((strcmp(dot, ".jpg")  == 0) ||
			(strcmp(dot, ".jpeg") == 0))
    {
		imgShape = GetImageShapeJpg(fileName);
	}
	else
    {
		CByteImage testImg;
		ImageIO::ReadFile(testImg, fileName);	
		imgShape = testImg.Shape();		
	}

	ENSURE(scale > 0.0f);
	imgShape.width  = (int) (imgShape.width  * scale);
	imgShape.height = (int) (imgShape.height * scale);

	return imgShape;
}

void ImageIO::WriteVideo(CFloatImage &video, const char* filePrefix)
{
	char frameName[1024];

	CShape videoShape = video.Shape();
	ENSURE(videoShape.depth > 1);
	for(int iFrame = 0; iFrame < videoShape.depth; iFrame++)
	{
		sprintf(frameName, "%s%05i.tga", filePrefix, iFrame);
		CFloatImage currFrame = video.GetFrame(iFrame);
		ImageIO::WriteFile(currFrame, frameName);
	}
}
