#include "Curve.h"

Curve::Curve()
{
	this->funcMap = NULL;
	this->funcSamples = 0;

	this->minOut = 0.0;
	this->maxOut = 1.0;

	CurvePoint leftPt(0.0, 0.0);	
	CurvePoint rightPt(1.0, 1.0);
	this->insert(leftPt);
	this->insert(rightPt);
}

Curve::Curve(const Curve& srcCurve)
{
	this->funcMap = NULL;
	*this = srcCurve;
}

Curve& Curve::operator= (const Curve& srcCurve)
{
	if(this != &srcCurve)
	{
		this->minOut      = srcCurve.minOut;
		this->maxOut      = srcCurve.maxOut;
		this->points      = srcCurve.points;
		this->funcSamples = srcCurve.funcSamples;

		if(this->funcMap)
			delete[] this->funcMap;

		this->funcMap = new double[this->funcSamples + 1];
		INSIST(this->funcMap != NULL);
		memcpy(this->funcMap, srcCurve.funcMap, (this->funcSamples + 1) * sizeof(*this->funcMap));
	}

	return *this;		
}

Curve::~Curve()
{
	if(this->funcMap != NULL)
		delete [] this->funcMap;
}

int Curve::insert(CurvePoint point, int _funcSamples)
{
	CHECK_RANGE(point.x, 0.0, 1.0);
	CHECK_RANGE(point.y, 0.0, 1.0);

	int iPoint;
	for(iPoint = 0; iPoint < (int) this->points.size(); iPoint++)
	{
		if(this->points[iPoint].x > point.x)
		{
			break;
		}
	}

	this->points.insert(this->points.begin() + iPoint, point);
	if(this->size() >= 2)
		ComputeFuncMap(_funcSamples);
	return iPoint;
}

void Curve::modify(int index, CurvePoint point)
{
	CHECK_RANGE(index, 0, this->size() - 1);
	CHECK_RANGE(point.x, 0.0, 1.0);
	CHECK_RANGE(point.y, 0.0, 1.0);
	if(index > 0)
	{
		INSIST(point.x >= this->points[index - 1].x)
	}
	if(index < this->size() - 1)
	{
		INSIST(point.x <= this->points[index + 1].x)
	}

	this->points[index] = point;
	ComputeFuncMap();
}

void Curve::erase(int ptIndex)
{
	CHECK_RANGE(ptIndex, 0, this->size() - 1);
	INSIST(this->size() > 2);	

	this->points.erase(this->points.begin() + ptIndex);	
	ComputeFuncMap();
}

int Curve::GetNearestPointIndex(double x, double y, double dist_thres)
{
	int near_pt_index = -1;
	double near_pt_dist_sqr = DBL_MAX;
	double dist_thres_sqr = dist_thres * dist_thres;

	for(uint iPoint = 0; iPoint < this->points.size(); iPoint++)
	{
		CurvePoint curvePoint = this->points[iPoint];
		double diffX = curvePoint.x - x;
		double diffY = curvePoint.y - y;
		double dist_sqr = (diffX * diffX) + (diffY * diffY);

		if((dist_sqr <= dist_thres_sqr) &&
		   (dist_sqr < near_pt_dist_sqr))
		{
			near_pt_dist_sqr = dist_sqr;
			near_pt_index = iPoint;
		}

		if((curvePoint.x > x) &&
		   ((curvePoint.x - x) > dist_thres))
		{
			break;
		}
	}

	return near_pt_index;
}

double Curve::GetCurveResponse(const void *curvePtr,  double x) 
{
	const Curve *curve = (Curve*) curvePtr;
	return curve->GetCurveResponse(x);
}

double Curve::GetCurveResponse(double x) const
{
	return GetCurveResponseFast(x);
	//return GetCurveResponsePrecise(x);
}

double Curve::GetCurveResponseFast(double x) const
{
	CHECK_RANGE(x, 0.0, 1.0);
	INSIST(this->funcSamples > 0);	

	int funcIndex = (int) (x * this->funcSamples);
	return this->funcMap[funcIndex];
}

void Curve::MinOut(double _minOut)
{
	INSIST(_minOut < this->maxOut);
	adjustResponseRange(_minOut, this->maxOut);
}

void Curve::MaxOut(double _maxOut)
{
	INSIST(_maxOut > this->minOut);
	adjustResponseRange(this->minOut, _maxOut);
}

void Curve::adjustResponseRange(double newMin, double newMax)
{
	double newSigRange = newMax - newMin;
	INSIST(newSigRange > 0.0);

	double oldSigRange = this->maxOut - this->minOut;
	INSIST(oldSigRange > 0.0);

	for(uint iPoint = 0; iPoint < this->points.size(); iPoint++)
	{
		double oldOut = (this->points[iPoint].y * oldSigRange) + this->minOut;  
		SET_IF_LT(oldOut, newMax);
		SET_IF_GT(oldOut, newMin);
		this->points[iPoint].y = (oldOut - newMin) / newSigRange;
	}

	this->minOut = newMin;
	this->maxOut = newMax;

	ComputeFuncMap();
}


void Curve::Dump()
{
	PRINT_INT(this->points.size());
	for(uint iPoint = 0; iPoint < this->points.size(); iPoint++)
	{
		printf("\t [%0.02f, %0.02f]\n", this->points[iPoint].x, this->points[iPoint].y);
	}
}

