#include "GS_FuncVis.h"
#include <FL/fl_draw.H>

void GS_FuncVis::getAxisLines(GS_Line &xAxis, GS_Line &yAxis,
							  const GS_FuncVisParams &params)
{
	yAxis.x1 = xPaddingLeft + axisCross;
	yAxis.y1 = yPaddingTop; 
	yAxis.x2 = yAxis.x1; 
	yAxis.y2 = params.h - yPaddingBottom; 
	
	xAxis.x1 = xPaddingLeft;
	xAxis.y1 = yAxis.y2 - axisCross; 
	xAxis.x2 = params.w - xPaddingRight;
	xAxis.y2 = xAxis.y1; 
}

GS_Line GS_FuncVis::GetGraphDiag(const GS_FuncVisParams &params)
{
	GS_Line graphDiag;

	if((params.w == 0) ||
	   (params.h == 0))
	{
		return graphDiag;		
	}

	GS_Line xAxis, yAxis;
	getAxisLines(xAxis, yAxis, params);	

	return getGraphDiag(xAxis, yAxis, params);
}

GS_Line GS_FuncVis::getGraphDiag(const GS_Line &xAxis, const  GS_Line &yAxis,
							     const GS_FuncVisParams &params)
{
	GS_Line graphDiag;
	graphDiag.x1 = params.x + yAxis.x1;
	graphDiag.y1 = params.y + xAxis.y1;
	graphDiag.x2 = params.x + xAxis.x2;
	graphDiag.y2 = params.y + yAxis.y1;
	return graphDiag;
}

GS_Line GS_FuncVis::draw(const GS_FuncVisParams &params)
{
	if((params.w == 0) ||
	   (params.h == 0))
	{
		return GS_Line();		
	}

	fl_push_clip(params.x, params.y, params.w, params.h);
	fl_push_matrix();
	fl_translate(params.x, params.y);

	GS_Line xAxis, yAxis;
	getAxisLines(xAxis, yAxis, params);	

	drawGrid(xAxis.y1, yAxis.x1, params);

	drawAxis(xAxis, yAxis);

	drawGraphDiag(xAxis, yAxis);

	drawFuncPlot(xAxis, yAxis, params);

	fl_pop_matrix();
	fl_pop_clip();

	drawAxisLabels(xAxis, yAxis, params);

	return getGraphDiag(xAxis, yAxis, params);
}

//draw the identity line
void GS_FuncVis::drawGraphDiag(const GS_Line &xAxis, const GS_Line &yAxis)
{
	Fl_Color incomingCol = fl_color();
	fl_color(FL_DARK1);

	//draw diagonal
	fl_begin_line();
	fl_vertex(yAxis.x1, xAxis.y1);
	fl_vertex(xAxis.x2, yAxis.y1);
	fl_end_line();

	fl_color(incomingCol);
}

//draw background grid lines
void GS_FuncVis::drawGrid(int xAxisY, int yAxisX, const GS_FuncVisParams &params)
{
	Fl_Color incomingCol = fl_color();
	fl_color(FL_LIGHT2);

	int gridSize = 10;	
	int maxGridLineCount = max((params.w / gridSize), (params.h / gridSize));
	for(int iGridLine = -maxGridLineCount; iGridLine < maxGridLineCount; iGridLine++)
	{		
		int vertGridLineX = yAxisX + (iGridLine * gridSize);
		if(InRange(vertGridLineX, 0, params.w))
		{
			fl_begin_line();
			fl_vertex(vertGridLineX, 0);
			fl_vertex(vertGridLineX, params.h);
			fl_end_line();
		}

		int horzGridLineY = xAxisY + (iGridLine * gridSize);
		if(InRange(horzGridLineY, 0, params.h))
		{
			fl_begin_line();		
			fl_vertex(0, horzGridLineY);
			fl_vertex(params.w, horzGridLineY);
			fl_end_line();
		}
	}

	fl_color(incomingCol);
}

void GS_FuncVis::drawAxis(const GS_Line &xAxis, const GS_Line &yAxis)
{
	//draw x-axis
	fl_begin_line();
	fl_vertex(yAxis.x1, yAxis.y1);
	fl_vertex(yAxis.x2, yAxis.y2);
	fl_end_line();

	//draw y-axis
	fl_begin_line();
	fl_vertex(xAxis.x1, xAxis.y1);
	fl_vertex(xAxis.x2, xAxis.y2);
	fl_end_line();
}

void GS_FuncVis::drawAxisLabels(const GS_Line &xAxis, const GS_Line &yAxis,
								const GS_FuncVisParams &params)
{
	int fontHeight = fl_height();
	char labelBuf[100];

	sprintf(labelBuf, "%0.01f", params.maxY);
	fl_draw(labelBuf, 
		    params.x, params.y + yAxis.y1, 
			xAxis.x1, fontHeight, FL_ALIGN_CENTER);

	sprintf(labelBuf, "%0.01f", params.minY);
	fl_draw(labelBuf, 
		    params.x, params.y + xAxis.y1 - (fontHeight / 2), 
		    xAxis.x1, fontHeight, FL_ALIGN_CENTER);

	int fontWidth;
	sprintf(labelBuf, "%0.01f", params.minX);
	fontWidth = (int) fl_width(labelBuf);
	fl_draw(labelBuf, 
	  	    params.x + yAxis.x1 - (fontWidth / 2), params.y + yAxis.y2,
			fontWidth, params.h - yAxis.y2, FL_ALIGN_CENTER);

	sprintf(labelBuf, "%0.01f", params.maxX);
	fontWidth = (int) fl_width(labelBuf);
	fl_draw(labelBuf, 
			params.x + xAxis.x2 - fontWidth, params.y + yAxis.y2,
			fontWidth, params.h - yAxis.y2, FL_ALIGN_CENTER);
}


void GS_FuncVis::drawFuncPlot(const GS_Line &xAxis, const GS_Line &yAxis, 
							  const GS_FuncVisParams &params)
{
	Fl_Color incomingCol = fl_color();
	fl_color(FL_BLUE);

	Callback_Double funcEvaluator = params.funcEvaluator;

	int xCoordStart = yAxis.x1;
	int xCoordEnd   = xAxis.x2;	
	int xCoordSpan  = xCoordEnd - xCoordStart;
	INSIST(xCoordSpan >= 0);

	int yCoordStart = yAxis.y1;
	int yCoordEnd   = xAxis.y1;	
	int yCoordSpan  = yCoordEnd - yCoordStart;
	INSIST(yCoordSpan >= 0);

	double xSignalSpan = params.maxX - params.minX;
	double ySignalSpan = params.maxY - params.minY;
	INSIST(xSignalSpan > 0);
	INSIST(ySignalSpan > 0);

	fl_begin_points();
	double xCoordIncr = 1.0;
	if(yCoordSpan > xCoordSpan)
	{
		xCoordIncr = (double) xCoordSpan / yCoordSpan;
	}
	xCoordIncr /= 10;
	for(double xCoord = xCoordStart; xCoord <= xCoordEnd; xCoord += xCoordIncr)
	{
		double xValNorm = (xCoord - xCoordStart) / xCoordSpan;
		double xVal = params.minX + (xSignalSpan * xValNorm);
		double yVal = funcEvaluator.second(funcEvaluator.first, xVal);;
		CHECK_RANGE(yVal, params.minY, params.maxY);

		double yValNorm = (yVal - params.minY) / ySignalSpan; 
		double yCoord   = yValNorm * yCoordSpan;
		yCoord = yCoordEnd - yCoord;

		fl_vertex(xCoord, yCoord);
	}
	fl_end_points();

	fl_color(incomingCol);
}
